/*
Module : MainFrm.h
Purpose: Defines the MFC frame class for EZView

Copyright (c) 2000 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise)
when your product is released in binary form. You are allowed to modify the source code in any way you want
except you cannot modify the copyright details at the top of each module. If you want to distribute source
code with your application, then you are only allowed to distribute versions released by the author. This is
to maintain a single distribution point for the source code.

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#pragma once


//////////////////// Classes //////////////////////////////////////////////////

class CMainFrame : public CFrameWndEx
{
public:
//Constructors / Destructors
  CMainFrame(const CMainFrame&) = delete;
  CMainFrame(CMainFrame&&) = delete;
  ~CMainFrame() = default; //NOLINT(modernize-use-override)

//Methods
  CMainFrame& operator=(const CMainFrame&) = delete;
  CMainFrame& operator=(CMainFrame&&) = delete;
  _NODISCARD bool FullScreen() const noexcept { return m_bFullScreen; };

protected:
  CMainFrame() noexcept; //create from serialization only

//Methods
  BOOL PreTranslateMessage(MSG* pMsg) override;

//Member variables
  CMFCMenuBar m_wndMenuBar;
  CMFCStatusBar m_wndStatusBar;
  CMFCToolBar m_wndToolBar;
  bool m_bFullScreen;
  CRect m_FullScreenWindowRect;
  WINDOWPLACEMENT m_wpPrev;

//Message handlers
  afx_msg int OnCreate(LPCREATESTRUCT lpCreateStruct);
  afx_msg void OnGetMinMaxInfo(MINMAXINFO* lpMMI);
  afx_msg void OnUpdateViewFullscreen(CCmdUI* pCmdUI);
  afx_msg void OnViewFullscreen();
  afx_msg void OnUpdateImage(CCmdUI* pCmdUI);
  afx_msg void OnViewCustomize();

  DECLARE_DYNCREATE(CMainFrame)

  DECLARE_MESSAGE_MAP()
};
